from typing import NotRequired, TypedDict, Required, Sequence, Literal, Any

from pydantic import ConfigDict


URL = str
Context = dict[str, Any]
Options = dict[str, Any]
ScrapeResult = dict[str, Any]

SchemaFieldType = Literal[
    "string",
    "str",
    "boolean",
    "bool",
    "integer",
    "int",
    "number",
    "float",
    "double",
    "currency",
    "email",
    "enum",
    "array",
    "object",
    "datetime",
    "phone_number",
    "url",
]


class Schema(TypedDict, total=False):
    __config__: NotRequired[ConfigDict]
    type: Required[SchemaFieldType]
    description: NotRequired[str]
    properties: NotRequired[dict[str, "Schema"]]
    items: NotRequired["Schema"]
    variants: NotRequired[Sequence[str]]
    expression: NotRequired[str]
    required: NotRequired[bool]


Schema.__annotations__["__extra_fields__"] = dict[str, "Schema"]


class Cookie(TypedDict):
    name: str
    value: str
    domain: str
    path: str
    expires: int | float
    size: int
    httpOnly: bool
    secure: bool
    session: bool
    sameSite: str


class LocalStorage(TypedDict):
    domain: str
    path: str | None
    key: str
    value: str



from .types import DownloadMeta, HTMLMetadata, ObservationTrigger
from .base import OutputObserver
from .logging_observer import LoggingObserver
from .memory_observer import InMemoryObserver
from .storage_observer import LocalStorageObserver

__all__ = [
    "DownloadMeta",
    "HTMLMetadata",
    "ObservationTrigger",
    "OutputObserver",
    "LoggingObserver",
    "InMemoryObserver",
    "LocalStorageObserver",
]



from typing import Any, List, Optional, Self, Type

from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    ValidationError,
    create_model,
    model_validator,
)

from harambe_core.errors import SchemaValidationError
from harambe_core.parser.constants import RESERVED_PREFIX
from harambe_core.parser.expression import ExpressionEvaluator
from harambe_core.parser.type_currency import ParserTypeCurrency
from harambe_core.parser.type_date import ParserTypeDate
from harambe_core.parser.type_email import ParserTypeEmail
from harambe_core.parser.type_enum import ParserTypeEnum
from harambe_core.parser.type_number import ParserTypeNumber
from harambe_core.parser.type_phone_number import ParserTypePhoneNumber
from harambe_core.parser.type_string import ParserTypeString
from harambe_core.parser.type_url import ParserTypeUrl
from harambe_core.types import Schema, SchemaFieldType


class SchemaParser:
    """
    A schema parser that uses Pydantic models to validate data against a JSON schema
    """

    def __init__(
        self, schema: Schema, evaluator: ExpressionEvaluator | None = None
    ) -> None:
        self._pk_expression = schema.get("$primary_key", None)

        if "$schema" in schema:
            del schema["$schema"]
        if "$primary_key" in schema:
            del schema["$primary_key"]

        if evaluator is None:
            evaluator = ExpressionEvaluator()

        self.evaluator = evaluator
        self.schema = schema
        self.field_types: dict[SchemaFieldType, Any] = {}

    def validate(self, data: dict[str, Any], base_url: str) -> dict[str, Any]:
        # Set these values here for convenience to avoid passing them around. A bit hacky
        self.field_types = self._get_field_types(base_url)
        model = self._schema_to_pydantic_model(self.schema, is_root=True)

        try:
            res = model(**data).model_dump(by_alias=True)
            if self._pk_expression:
                res["$primary_key"] = self.evaluator.evaluate(self._pk_expression, res)
            return res

        except (TypeError, ValidationError) as e:
            raise SchemaValidationError(message=str(e))

    @staticmethod
    def _get_field_types(base_url: str) -> dict[SchemaFieldType, Any]:
        return {
            "string": ParserTypeString,
            "str": ParserTypeString,
            "boolean": bool,
            "bool": bool,
            "integer": int,
            "int": int,
            "number": ParserTypeNumber,
            "float": ParserTypeNumber,
            "double": ParserTypeNumber,
            "currency": ParserTypeCurrency(),
            "email": ParserTypeEmail,
            "enum": ParserTypeEnum,
            "array": list,
            "object": dict[str, Any],
            "datetime": ParserTypeDate(),
            "phone_number": ParserTypePhoneNumber(),
            "url": ParserTypeUrl(base_url=base_url),
        }

    def _items_schema_to_python_type(
        self, items_info: Schema, model_name: str = "DynamicModelItem"
    ) -> Type[Any]:
        """
        Convert a JSON schema's items property to a Python type
        """
        item_type = items_info.get("type")
        if item_type is None:
            raise ValueError(f"Item type for array `{model_name}` is missing")

        if item_type == "object":
            return self._schema_to_pydantic_model(
                items_info.get("properties", {}),
                model_name=f"{model_name}Object",
                is_root=False,
            )

        if item_type == "array":
            return List[  # type: ignore
                self._items_schema_to_python_type(
                    items_info.get("items", {}),
                    model_name=f"{model_name}Item",
                )
            ]

        # Every enum has its own unique variants
        if item_type == "enum":
            return self._get_type(item_type, required=True)(items_info["variants"])

        return self._get_type(item_type, required=True)

    def _schema_to_pydantic_model(
        self,
        schema: Schema,
        model_name: str = "DynamicModel",
        required: bool = True,
        is_root: bool = True,
    ) -> Type[BaseModel]:
        """
        Convert a JSON schema to a Pydantic model dynamically. All fields are optional
        """
        fields = {}
        computed_fields = {}

        for field_name, field_info in schema.items():
            if field_name == "__config__":
                continue

            field_serialization_alias = None
            if field_name.startswith("model_"):
                field_serialization_alias = field_name
                field_name = RESERVED_PREFIX + field_name

            field_description = field_info.get("description", None)
            field_type = field_info.get("type")
            field_required = field_info.get("required", False)
            field = Field(
                ...,
                description=field_description,
                alias=field_serialization_alias,
                serialization_alias=field_serialization_alias,
                min_length=1 if field_type == "array" and field_required else None,
            )

            if field_type == "object":
                python_type = self._schema_to_pydantic_model(
                    field_info.get("properties", {}) or {},
                    model_name=f"{model_name}{field_name.capitalize()}",
                    required=field_required,
                    is_root=False,
                )
            elif field_type == "array":
                python_type = List[  # type: ignore
                    self._items_schema_to_python_type(
                        field_info.get("items", {}) or {},
                        model_name=f"{model_name}Item",
                    )
                ]
            elif field_type == "enum":
                python_type = self._get_type(field_type, required=True)(
                    *field_info["variants"]
                )
            elif expression := field_info.get("expression"):
                python_type = self._get_type(field_type, required=False)
                field = Field(
                    default=None,
                    description=field_description,
                    serialization_alias=field_serialization_alias,
                )
                computed_fields[field_name] = expression
            else:
                python_type = self._get_type(field_type, required=field_required)

            fields[field_name] = (python_type, field)

        config: ConfigDict = {"extra": "forbid", "str_strip_whitespace": True}
        config.update(schema.get("__config__", {}))
        base_model = base_model_factory(
            config, computed_fields, self.evaluator, is_root
        )
        new_model = create_model(model_name, __base__=base_model, **fields)

        if not required:
            new_model = Optional[new_model]

        return new_model

    def _get_type(self, field: SchemaFieldType, required: bool | None) -> Type[Any]:
        field_type = self.field_types.get(field)
        if not field_type:
            raise ValueError(f"Unsupported field type: {field}")
        if not required:
            field_type = Optional[field_type]
        return field_type


def base_model_factory(
    config: ConfigDict,
    computed_fields: dict[str, str],
    evaluator: ExpressionEvaluator,
    is_root: bool = True,
) -> Type[BaseModel]:
    class PreValidatedBaseModel(BaseModel):
        model_config: ConfigDict = config

        # noinspection PyNestedDecorators
        @model_validator(mode="before")
        @classmethod
        def pre_validate(cls, values: Any) -> Any:
            def trim_and_nullify(value: Any) -> Any:
                if isinstance(value, str):
                    value = value.strip()
                    if not value:
                        return None
                if isinstance(value, list):
                    return [trim_and_nullify(v) for v in value]
                if isinstance(value, dict):
                    return {k.strip(): trim_and_nullify(v) for k, v in value.items()}
                return value

            return trim_and_nullify(values)

        @model_validator(mode="after")
        def post_validate(self) -> Self:
            for field, expression in computed_fields.items():
                res = evaluator.evaluate(expression, self)
                setattr(self, field, res)

            # Only check for empty fields at the root level
            if is_root and _all_fields_empty(self.model_dump()):
                raise SchemaValidationError(
                    message=f"All fields are null or empty. data={self.model_dump()}",
                )

            return self

    return PreValidatedBaseModel


def _all_fields_empty(data: dict[str, Any]) -> bool:
    """
    Check if all fields in the base object are either None or empty recursively.
    This does not get called for individual fields in the base object.
    """

    def is_empty(value: Any) -> bool:
        if value is None:
            return True
        if isinstance(value, dict):
            return all(is_empty(v) for v in value.values())
        if isinstance(value, list):
            return all(is_empty(v) for v in value)
        if isinstance(value, str):
            return not value.strip()
        return False

    return all(is_empty(value) for value in data.values())



from pprint import pprint
from typing import List
from urllib.parse import quote

from harambe_core.types import Cookie, LocalStorage, URL, Context, Options, ScrapeResult
from .base import OutputObserver
from .types import DownloadMeta


class LoggingObserver(OutputObserver):
    async def on_save_data(self, data: ScrapeResult) -> None:
        pprint(data, width=240)

    async def on_queue_url(self, url: URL, context: Context, options: Options) -> None:
        print(f"Enqueuing: {url} with context {context} and options {options}")

    async def on_download(
        self, download_url: str, filename: str, content: bytes
    ) -> "DownloadMeta":
        print(f"Downloading file: {filename}")
        return {
            "url": f"{download_url}/{quote(filename)}",
            "filename": filename,
        }

    async def on_paginate(self, next_url: str) -> None:
        pass

    async def on_save_cookies(self, cookies: List[Cookie]) -> None:
        print(f"Cookies saved : {cookies}")

    async def on_save_local_storage(self, local_storage: List[LocalStorage]) -> None:
        print(f"Local Storage saved : {local_storage}")

    async def on_check_and_solve_captchas(self, page: "Page") -> None:
        pass



from typing import Any, List, Callable

from pydantic.functional_validators import AfterValidator
from typing_extensions import Annotated


class ParserTypeEnum:
    def __new__(cls, *variants: str) -> Any:
        return Annotated[str, AfterValidator(cls.validate_type(*variants))]

    @staticmethod
    def validate_type(*variants: str) -> Callable[[str], str]:
        variant_map = {v.strip().lower(): v for v in variants}

        def _validate_type(value: str) -> str:
            if (value := value.strip().lower()) not in variant_map:
                raise ValueError(
                    f'Value "{value}" doesn\'t match any of the variants ({variants})'
                )

            return variant_map[value]

        return _validate_type



import re
from typing import Any

from pydantic import BeforeValidator
from typing_extensions import Annotated

price_not_available_phrases = [
    "price not available",
    "unavailable price",
    "price upon request",
    "contact for price",
    "request a quote",
    "call for price",
    "check price in store",
    "price tbd (to be determined)",
    "price not disclosed",
    "out of stock",
    "sold out",
    "pricing not provided",
    "not priced",
    "currently unavailable",
    "n/a (not available)",
    "ask for pricing",
    "see details for price",
    "price coming soon",
    "temporarily unavailable",
    "price hidden",
    "tdb",
    "n/a",
]


class ParserTypeCurrency:
    def __new__(cls) -> Any:
        return Annotated[float | None, BeforeValidator(cls.validate_currency)]

    @staticmethod
    def validate_currency(value: str) -> float | None:
        if isinstance(value, (float, int)):
            return float(value)

        value = str(value).strip()

        if value.lower() in price_not_available_phrases:
            return None

        cleaned_value = re.sub(r"[^\d.,-]", "", value)
        cleaned_value = re.sub(r"^0+(?!$)", "", cleaned_value)

        if "." in cleaned_value:
            decimal_parts = cleaned_value.split(".")
            if len(decimal_parts[-1]) == 3:  # thousands separator issue
                cleaned_value = cleaned_value.replace(".", "")

        if cleaned_value.startswith("."):
            cleaned_value = "0" + cleaned_value
            return float(cleaned_value)

        if "," in cleaned_value and "." in cleaned_value:
            if cleaned_value.index(",") < cleaned_value.index("."):
                cleaned_value = cleaned_value.replace(",", "")
            else:
                cleaned_value = cleaned_value.replace(".", "").replace(",", ".")
        elif "," in cleaned_value and "." not in cleaned_value:
            if len(cleaned_value.split(",")[-1]) == 2:
                cleaned_value = cleaned_value.replace(",", ".")

            elif (
                len(cleaned_value.split(",")[-1]) != 3
            ):  # check Ambiguous values 123,45 and 123,456
                raise ValueError("Invalid price")
            cleaned_value = cleaned_value.replace(",", "")

        return float(cleaned_value.strip())



from typing import Any, Callable, TypeVar, TypedDict
from urllib.parse import quote

from harambe_core.observer import OutputObserver
from harambe_core.types import LocalStorage, Cookie, URL, Context, Options
from .types import DownloadMeta, ObservationTrigger


class Payload(TypedDict):
    type: ObservationTrigger
    data: dict[str, Any] | list[Cookie] | list[LocalStorage]


T = TypeVar("T")

Serializer = Callable[[Payload], T]
Sink = Callable[[T], None]


class SerializationObserver(OutputObserver):
    def __init__(self, sink: Sink, serializer: Serializer) -> None:
        self._publish: Callable[[Payload], None] = lambda x: sink(serializer(x))

    async def on_save_data(self, data: dict[str, Any]) -> None:
        payload: Payload = {"type": "on_save_data", "data": data}

        self._publish(payload)

    async def on_queue_url(self, url: URL, context: Context, options: Options) -> None:
        payload: Payload = {
            "type": "on_queue_url",
            "data": {"url": url, "context": context, "options": options},
        }

        self._publish(payload)

    async def on_download(
        self, download_url: str, filename: str, content: bytes
    ) -> "DownloadMeta":
        payload: Payload = {
            "type": "on_download",
            "data": {
                "download_url": download_url,
                "filename": filename,
                "content": content.decode("utf-8"),
            },
        }

        self._publish(payload)

        return {
            "url": f"{download_url}/{quote(filename)}",
            "filename": filename,
        }

    async def on_paginate(self, next_url: str) -> None:
        pass

    async def on_save_cookies(self, cookies: list[Cookie]) -> None:
        payload: Payload = {"type": "on_save_cookies", "data": cookies}

        self._publish(payload)

    async def on_save_local_storage(self, local_storage: list[LocalStorage]) -> None:
        payload: Payload = {"type": "on_save_local_storage", "data": local_storage}

        self._publish(payload)



from typing import Any, List, Tuple
from urllib.parse import quote

from harambe_core.types import URL, Context, Options, Cookie, LocalStorage
from .base import OutputObserver
from .types import DownloadMeta


class InMemoryObserver(OutputObserver):
    def __init__(self) -> None:
        self._data: List[dict[str, Any]] = []
        self._urls: List[Tuple[URL, Context, Options]] = []
        self._files: List[Tuple[str, bytes]] = []
        self._cookies: List[Cookie] = []
        self._local_storage: List[LocalStorage] = []

    async def on_save_data(self, data: dict[str, Any]) -> None:
        self._data.append(data)

    async def on_queue_url(self, url: URL, context: Context, options: Options) -> None:
        self._urls.append((url, context, options))

    async def on_download(
        self, download_url: str, filename: str, content: bytes
    ) -> "DownloadMeta":
        self._files.append((filename, content))
        return {
            "url": f"{download_url}/{quote(filename)}",
            "filename": filename,
        }

    async def on_paginate(self, next_url: str) -> None:
        pass

    async def on_save_cookies(self, cookies: list[Cookie]) -> None:
        self._cookies.extend(cookies)

    async def on_check_and_solve_captchas(self, page: "Page") -> None:
        pass

    async def on_save_local_storage(self, local_storage: list[LocalStorage]) -> None:
        self._local_storage.extend(local_storage)

    @property
    def data(self) -> List[dict[str, Any]]:
        return self._data

    @property
    def urls(self) -> List[Tuple[URL, Context, Options]]:
        return self._urls

    @property
    def files(self) -> List[Tuple[str, bytes]]:
        return self._files

    @property
    def cookies(self) -> List[Cookie]:
        return self._cookies

    @property
    def local_storage(self) -> List[LocalStorage]:
        return self._local_storage



from typing import Any, Generator

from slugify import slugify as python_slugify

from harambe_core.parser.expression.evaluator import ExpressionEvaluator


@ExpressionEvaluator.define_builtin("NOOP")
def noop(*args: Any) -> Any:
    return args[0] if len(args) == 1 else args


@ExpressionEvaluator.define_builtin("CONCAT")
def concat(*args: Any, seperator: str = "") -> str:
    args = [a for a in flatten(args) if a is not None and a != ""]
    return seperator.join(str(arg) for arg in args)


@ExpressionEvaluator.define_builtin("CONCAT_WS")
def concat_ws(seperator: str, *args: Any) -> str:
    return concat(*args, seperator=seperator)


@ExpressionEvaluator.define_builtin("COALESCE")
def coalesce(*args: Any) -> Any:
    for arg in args:
        if arg:
            return arg
    return None


@ExpressionEvaluator.define_builtin("SLUGIFY")
def slugify(*args: Any) -> str:
    text = concat_ws("-", *args)
    return python_slugify(text)


@ExpressionEvaluator.define_builtin("UPPER")
def upper(text: str) -> str:
    return text.upper()


@ExpressionEvaluator.define_builtin("LOWER")
def lower(text: str) -> str:
    return text.lower()


@ExpressionEvaluator.define_builtin("SUBSTRING_AFTER")
def substring_after(input_string: str, delimiter: str) -> str:
    """
    Returns the substring after the first occurrence of the specified delimiter.

    Args:
        input_string: The string to extract from
        delimiter: The substring to search for

    Returns:
        The substring after the first occurrence of the delimiter.
        If the delimiter is not found, returns the original string.
    """
    if delimiter == "":
        raise ValueError("SUBSTRING_AFTER requires a non-empty delimiter to search for")

    parts = input_string.split(delimiter, 1)
    if len(parts) == 1:
        return input_string

    return parts[1]


def flatten(values: Any) -> Generator[Any, None, None]:
    for value in values:
        if isinstance(value, list):
            yield from flatten(value)
        else:
            yield value



from abc import abstractmethod
from typing import runtime_checkable, Protocol, Any

from harambe_core.types import URL, Context, Options, Cookie, LocalStorage


@runtime_checkable
class OutputObserver(Protocol):
    @abstractmethod
    async def on_save_data(self, data: dict[str, Any]) -> None:
        raise NotImplementedError()

    @abstractmethod
    async def on_queue_url(self, url: URL, context: Context, options: Options) -> None:
        raise NotImplementedError()

    @abstractmethod
    async def on_download(
        self, download_url: str, filename: str, content: bytes
    ) -> "DownloadMeta":
        raise NotImplementedError()

    @abstractmethod
    async def on_paginate(self, next_url: str) -> None:
        raise NotImplementedError()

    @abstractmethod
    async def on_save_cookies(self, cookies: list[Cookie]) -> None:
        raise NotImplementedError()

    @abstractmethod
    async def on_save_local_storage(self, local_storage: list[LocalStorage]) -> None:
        raise NotImplementedError()
